#define com_maroontress_libcgi_db_BoxKit_IMPORT
#define com_maroontress_libcgi_db_Tree_IMPORT
#define com_maroontress_libcgi_db_XML_IMPORT
#define std_db_BTree_IMPORT
#define std_io_BufferedOutputTube_IMPORT
#define std_io_OutputTube_IMPORT
#define std_lang_ByteArray_IMPORT
#define std_lang_Chunk_IMPORT
#define std_lang_ImmutableChunk_IMPORT
#define std_util_Pool_IMPORT

#include <sys/types.h>
#include <sys/param.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <com/maroontress/libcgi/db/BoxKit.h>
#include <com/maroontress/libcgi/db/Tree.h>
#include <com/maroontress/libcgi/db/XML.h>
#include <std/db/BTree.h>
#include <std/io/BufferedOutputTube.h>
#include <std/io/OutputTube.h>
#include <std/lang/ByteArray.h>
#include <std/lang/Chunk.h>
#include <std/lang/ImmutableChunk.h>
#include <std/util/Pool.h>

struct BoxKit {
    struct Tree *tree;
    char *key;
    union BTree db;
};

struct BoxKit *
BoxKit_create(struct Tree *tree, const char *key)
{
    static const int MODE = 0666;
    static const int FLAGS = (O_CREAT | O_RDWR);
    struct BoxKit *kit;

    if ((kit = (struct BoxKit *)malloc(sizeof(*kit))) == NULL)
        return NULL;
    kit->tree = tree;
    kit->key = NULL;
    kit->db.impl = NULL;

    if ((kit->key = strdup(key)) == NULL
        || (kit->db = BTree_create(NULL, FLAGS, MODE)).impl == NULL) {
	BoxKit_purge(kit);
	return NULL;
    }
    return kit;
}

int
BoxKit_load(struct BoxKit *kit, const void *data, size_t size)
{
    const char *d = data;
    char *z;
    size_t n;
    struct ImmutableChunk key, val;

    while (size > 0) {
	if (size < sizeof(n))
	    break;
        bcopy(d, &n, sizeof(n));
        d += sizeof(n);
	size -= sizeof(n);

	if (n == 0 || n > size || (z = memchr(d, 0, n)) == NULL)
	    break;
        key.data = d;
        key.size = z - d + 1;
        d += key.size;

        val.data = d;
        val.size = n - key.size;
        d += val.size;

        if (BTree_put(kit->db, key, val) < 0)
	    break;
        size -= n;
    }
    return (size > 0) ? -1 : 0;
}

int
BoxKit_setInt(struct BoxKit *kit, const char *name, int value)
{
    struct ImmutableChunk key, val;
    char data[1 + sizeof(int)];

    key.data = name;
    key.size = strlen(name) + 1;
    data[0] = 'i';
    bcopy(&value, data + 1, sizeof(int));
    val.data = data;
    val.size = sizeof(data);
    return BTree_put(kit->db, key, val);
}

int
BoxKit_getInt(struct BoxKit *kit, const char *name, /*@out@*/ int *value)
{
    struct ImmutableChunk key;
    struct Chunk v;

    key.data = name;
    key.size = strlen(name) + 1;
    if (BTree_get(kit->db, key, &v) != 0
	|| v.size != 1 + sizeof(int)
	|| *(const char *)v.data != 'i') {
        return -1;
    }
    bcopy((const char *)v.data + 1, value, sizeof(*value));
    return 0;
}

int
BoxKit_setString(struct BoxKit *kit, const char *name, const char *value)
{
    struct ImmutableChunk key, val;
    size_t size;
    char *data;

    key.data = name;
    key.size = strlen(name) + 1;
    size = strlen(value) + 1;
    if ((data = alloca(1 + size)) == NULL) {
        return -1;
    }
    data[0] = 's';
    bcopy(value, data + 1, size);
    val.data = data;
    val.size = 1 + size;
    return BTree_put(kit->db, key, val);
}

static int
createDuplicatedString(struct BoxKit *kit, char *(*duplicate)(const char *),
		       const char *name, /*@out@*/ char **value)
{
    struct ImmutableChunk key;
    struct Chunk v;

    key.data = name;
    key.size = strlen(name) + 1;
    if (BTree_get(kit->db, key, &v) != 0
	|| v.size < 2
	|| *(const char *)v.data != 's'
	|| ((const char *)v.data)[v.size - 1] != 0
	|| (*value = duplicate((const char *)v.data + 1)) == NULL) {
        return -1;
    }
    return 0;
}

int
BoxKit_createString(struct BoxKit *kit, const char *name,
		    /*@out@*/ char **value)
{
    return createDuplicatedString(kit, strdup, name, value);
}

int
BoxKit_createXMLEscapedString(struct BoxKit *kit, const char *name,
			      /*@out@*/ char **value)
{
    return createDuplicatedString(kit, XML_createEscaped, name, value);
}

int
BoxKit_delete(struct BoxKit *kit, const char *name)
{
    struct ImmutableChunk key;

    key.data = name;
    key.size = strlen(name) + 1;
    return BTree_del(kit->db, key);
}

char *
BoxKit_getKey(struct BoxKit *kit)
{
    return kit->key;
}

void
BoxKit_purge(struct BoxKit *kit)
{
    if (kit == NULL)
	return;
    free(kit->key);
    if (kit->db.impl != NULL)
	BTree_close(kit->db);
    free(kit);
}

static int
fill(struct BoxKit *kit, union OutputTube out)
{
    struct Chunk k, v;
    int s, size;
    
    for (s = BTree_seq(kit->db, &k, &v, BTree_FIRST); s == 0;
         s = BTree_seq(kit->db, &k, &v, BTree_NEXT)) {
        size = k.size + v.size;
	if (OutputTube_writeFully(out, &size, sizeof(size)) < 0
	    || OutputTube_writeFully(out, k.data, k.size) < 0
	    || OutputTube_writeFully(out, v.data, v.size) < 0) {
	    return -1;
	}
    }
    return (s < 0) ? -1 : 0;
}

static int
fillPool(struct BoxKit *kit, union Pool p)
{
    union BufferedOutputTube out;
    int s;
    
    if ((out = BufferedOutputTube_create(Pool_createOutputTube(p))).impl
	== NULL) {
	return -1;
    }
    s = fill(kit, out.super.outputTube);
    return (OutputTube_close(out.super.outputTube) < 0) ? -1 : s;
}

static int
commitWithPool(struct BoxKit *kit, union Pool p)
{
    struct ByteArray *a;
    struct Chunk c;
    int s;

    if (fillPool(kit, p) < 0 || (a = Pool_createByteArray(p)) == NULL) {
	return -1;
    }
    c = ByteArray_getChunk(a);
    s = Tree_put(kit->tree, kit->key, c.data, c.size);
    ByteArray_free(a);
    return s;
}

static int
commit(struct BoxKit *kit)
{
    union Pool p;
    int s;

    if ((p = Pool_create()).impl == NULL)
	return -1;
    s = commitWithPool(kit, p);
    Pool_free(p);
    return s;
}

int
BoxKit_close(struct BoxKit *kit)
{
    int s;
	
    s = commit(kit);
    BoxKit_purge(kit);
    return s;
}
