#define com_maroontress_libcgi_db_Box_IMPORT
#define com_maroontress_libcgi_db_BoxKit_IMPORT
#define com_maroontress_libcgi_db_Tree_IMPORT
#define std_db_BTree_IMPORT
#define std_lang_Chunk_IMPORT
#define std_lang_ImmutableChunk_IMPORT

#include <sys/types.h>
#include <sys/param.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <com/maroontress/libcgi/db/Box.h>
#include <com/maroontress/libcgi/db/BoxKit.h>
#include <com/maroontress/libcgi/db/Tree.h>
#include <std/db/BTree.h>
#include <std/lang/Chunk.h>
#include <std/lang/ImmutableChunk.h>

#define HAS_BSD_OPEN (!defined(_POSIX_SOURCE)				\
                      && defined(O_SHLOCK) && defined(O_EXLOCK))
#if !HAS_BSD_OPEN
#include <sys/file.h>
#define O_EXLOCK 0
#define O_SHLOCK 0
#endif

struct Tree {
    const struct Tree_Mode *mode;
    union BTree db;
    struct Box *box;
    int hasError;
};

struct Tree_Mode {
    int flags;
    int mode;
    int operation;
};

static int
syncTree(struct Tree *t)
{
    if (BTree_sync(t->db) < 0) {
	t->hasError = 1;
	return -1;
    }
    return 0;
}

#define FLAGS_RO (O_RDONLY | O_SHLOCK)
#define FLAGS_SH (O_CREAT | O_RDONLY | O_SHLOCK)
#define FLAGS_EX (O_CREAT | O_RDWR | O_EXLOCK)

static const struct Tree_Mode MODE_READONLY = {FLAGS_RO, 0, LOCK_SH};
const struct Tree_Mode *const Tree_MODE_READONLY = &MODE_READONLY;

static const struct Tree_Mode MODE_SHAERD = {FLAGS_SH, 0666, LOCK_SH};
const struct Tree_Mode *const Tree_MODE_SHARED = &MODE_SHAERD;

static const struct Tree_Mode MODE_EXCLUSIVE = {FLAGS_EX, 0666, LOCK_EX};
const struct Tree_Mode *const Tree_MODE_EXCLUSIVE = &MODE_EXCLUSIVE;

struct Tree *
Tree_create(const char *path, const struct Tree_Mode *m)
{
    struct Tree *t;

    if ((t = (struct Tree *)malloc(sizeof(*t))) == NULL)
	return NULL;
    t->mode = m;
    t->db.impl = NULL;
    t->box = NULL;
    t->hasError = 0;

    if ((t->db = BTree_create(path, m->flags, m->mode)).impl == NULL) {
	Tree_free(t);
	return NULL;
    }
#if !HAS_BSD_OPEN
    flock(BTree_getFileDescriptor(t->db), m->operation);
#endif
    return t;
}

void
Tree_free(struct Tree *t)
{
    if (t == NULL)
	return;
    if (t->db.impl != NULL)
	BTree_close(t->db);
    Box_free(t->box);
    free(t);
}

struct BoxKit *
Tree_getBoxKit(struct Tree *t, const char *name)
{
    struct BoxKit *kit;
    struct ImmutableChunk key;
    struct Chunk v;
    int s;

    key.data = name;
    key.size = strlen(name) + 1;
    if ((s = BTree_get(t->db, key, &v)) < 0
	|| (kit = BoxKit_create(t, name)) == NULL) {
	t->hasError = 1;
	return NULL;
    }
    if (s == 0 && BoxKit_load(kit, v.data, v.size) < 0) {
	BoxKit_purge(kit);
	t->hasError = 1;
	return NULL;
    }
    return kit;
}

int
Tree_delete(struct Tree *t, const char *name)
{
    struct ImmutableChunk key;
    int s;

    key.data = name;
    key.size = strlen(name) + 1;
    if ((s = BTree_del(t->db, key)) < 0) {
	t->hasError = 1;
	return -1;
    }
    if (s == 1) {
	return 1;
    }
    return syncTree(t);
}

static struct Box *
updateBox(struct Tree *t, const char *key, const void *data, size_t size)
{
    Box_free(t->box);
    if ((t->box = Box_create(key, data, size)) == NULL) {
	t->hasError = 1;
	return NULL;
    }
    return t->box;
}

struct Box *
Tree_getBox(struct Tree *t, const char *name)
{
    struct ImmutableChunk key;
    struct Chunk v;
    int s;

    key.data = name;
    key.size = strlen(name) + 1;
    if ((s = BTree_get(t->db, key, &v)) < 0) {
	t->hasError = 1;
	return NULL;
    }
    if (s == 1) {
	return NULL;
    }
    return updateBox(t, name, v.data, v.size);
}

static struct Box *
scanSequentialBox(struct Tree *t, const struct BTree_SeqOption *opt)
{
    struct Chunk k, v;
    int s;

    if ((s = BTree_seq(t->db, &k, &v, opt)) < 0) {
	t->hasError = 1;
	return NULL;
    }
    if (s == 1) {
	return NULL;
    }
    if (memchr(k.data, 0, k.size) == NULL) {
	t->hasError = 1;
	return NULL;
    }
    return updateBox(t, k.data, v.data, v.size);
}

struct Box *
Tree_scanFirstBox(struct Tree *t)
{
    return scanSequentialBox(t, BTree_FIRST);
}

struct Box *
Tree_scanNextBox(struct Tree *t)
{
    return scanSequentialBox(t, BTree_NEXT);
}

struct Box *
Tree_scanLastBox(struct Tree *t)
{
    return scanSequentialBox(t, BTree_LAST);
}

struct Box *
Tree_scanPreviousBox(struct Tree *t)
{
    return scanSequentialBox(t, BTree_PREV);
}

struct Box *
Tree_scanBox(struct Tree *t, const char *name)
{
    struct ImmutableChunk key;
    struct Chunk k, v;
    int s;

    key.data = name;
    key.size = strlen(name) + 1;
    if ((s = BTree_seqCursor(t->db, key, &k, &v)) < 0) {
	t->hasError = 1;
	return NULL;
    }
    if (s == 1) {
	return NULL;
    }
    if (memchr(k.data, 0, k.size) == NULL) {
	t->hasError = 1;
	return NULL;
    }
    return updateBox(t, k.data, v.data, v.size);
}

int
Tree_put(struct Tree *t, const char *name, const void *data, size_t size)
{
    struct ImmutableChunk key, val;

    key.data = name;
    key.size = strlen(name) + 1;
    val.data = data;
    val.size = size;
    if (BTree_put(t->db, key, val) < 0) {
	t->hasError = 1;
	return -1;
    }
    return syncTree(t);
}

int
Tree_hasError(struct Tree *t)
{
    return t->hasError;
}

void
Tree_clearError(struct Tree *t)
{
    t->hasError = 0;
}
